import torch 
from tensornets import CTMRG

'''
customize the backward of ctmrg contraction
we simply provide the adjoint of T so it does not need to diff into C and E
however, since we have enable_grad in the backward, the double backward will enter C and E
'''
class Contraction(torch.autograd.Function):
    @staticmethod
    def forward(self, T, chi, maxiter, epsilon, use_checkpoint):

        self.chi = chi 
        self.maxiter = maxiter 
        self.epsilon = epsilon 
        self.use_checkpoint = use_checkpoint 

        C, E = CTMRG(T, chi, maxiter, epsilon, use_checkpoint=use_checkpoint)

        Z1 = torch.einsum('ab,bcd,fd,gha,chij,fjk,lg,mil,mk', (C,E,C,E,T,E,C,E,C))
        Z3 = torch.einsum('ab,bc,cd,da', (C,C,C,C))
        Z2 = torch.einsum('ab,bcd,de,fa,gcf,ge',(C,E,C,C,E,C))
        
        self.save_for_backward(T)
        return torch.log(Z1.abs()) + torch.log(Z3.abs()) - 2.*torch.log(Z2.abs())

    @staticmethod
    def backward(self, dlnZ):   
        T, = self.saved_tensors
        with torch.enable_grad():
            C, E = CTMRG(T, self.chi, self.maxiter, self.epsilon, use_checkpoint=self.use_checkpoint)

        up = torch.einsum('ab,bcd,fd,gha,fjk,lg,mil,mk->chij', (C,E,C,E,E,C,E,C)) * dlnZ
        dn = torch.einsum('ab,bcd,fd,gha,chij,fjk,lg,mil,mk', (C,E,C,E,T,E,C,E,C))
        return up/dn, None, None, None, None

def symmetrize(A):
    '''
    A(phy, up, left, down, right)
    left-right, up-down, diagonal symmetrize
    '''
    Asymm = (A + A.permute(0, 1, 4, 3, 2))/2.           # left-right symmetry
    Asymm = (Asymm + Asymm.permute(0, 3, 2, 1, 4))/2.   # up-down symmetry
    Asymm = (Asymm + Asymm.permute(0, 4, 3, 2, 1))/2.   # skew-diagonal symmetry
    Asymm = (Asymm + Asymm.permute(0, 2, 1, 4, 3))/2.   # diagonal symmetry

    return Asymm/Asymm.norm()


if __name__=='__main__':
    import numpy as np
    d = 2 
    D = 2
    chi = 50 
    maxiter = 50 
    epsilon = 1E-10
    dtype = torch.float64

    #dimer covering
    T = torch.zeros(d, d, d, d, d, d, dtype=dtype)
    T[0, 0, 0, 0, 0, 1] = 1.0 
    T[0, 0, 0, 0, 1, 0] = 1.0 
    T[0, 0, 0, 1, 0, 0] = 1.0 
    T[0, 0, 1, 0, 0, 0] = 1.0 
    T[0, 1, 0, 0, 0, 0] = 1.0 
    T[1, 0, 0, 0, 0, 0] = 1.0 
    T = T.view(d, d**4, d)
    
    contract = Contraction.apply

    def f(x):
        A = torch.as_tensor(x).view(d, D, D, D, D)
        As = symmetrize(A).view(d, D**4)
        T2 = (As.t()@As).view(D, D, D, D, D, D, D, D).permute(0,4, 1,5, 2,6, 3,7).contiguous().view(D**2, D**2, D**2, D**2)
        return -contract(T2, chi, maxiter, epsilon) 
    
    def g(x):
        A = torch.as_tensor(x).view(d, D, D, D, D).requires_grad_()
        As = symmetrize(A).view(d, D**4)
        T2 = (As.t()@As).view(D, D, D, D, D, D, D, D).permute(0,4, 1,5, 2,6, 3,7).contiguous().view(D**2, D**2, D**2, D**2)
        loss = -contract(T2, chi, maxiter, epsilon) 
        loss.backward()
        return (A.grad).numpy().ravel()

    from scipy.optimize import check_grad
    x = np.random.randn(d*D**4)
    print (f(x), g(x).shape)
    print ('gradient check:',  check_grad(f, g, x))
